package org.prlab.idic.privacypolicy;

import android.bluetooth.*;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.os.Bundle;
import android.support.v7.app.AppCompatActivity;
import android.support.v7.widget.RecyclerView;
import android.util.Log;
import android.view.View;
import android.widget.*;
import org.prlab.idic.privacypolicy.adapter.MessageRecyclerViewAdapter;
import org.prlab.idic.privacypolicy.ble.BLECustomScript;
import org.prlab.idic.privacypolicy.pojo.Message;
import org.prlab.idic.privacypolicy.schedule.Scheduled;
import org.prlab.idic.privacypolicy.storage.PrivacyPolicyStorageImpl;
import org.prlab.idic.privacypolicy.tool.BaseUtils;

import java.util.LinkedList;
import java.util.List;
import java.util.UUID;

import static android.bluetooth.BluetoothGatt.GATT_SUCCESS;

public class DeviceOperatorActivity extends AppCompatActivity {

    private static final String TAG = DeviceOperatorActivity.class.getSimpleName();
    private UUID PRIVACY_POLICY_ACCEPT_UUID;

    private PrivacyPolicyStorageImpl mPolicyStorage;
    private String mTargetDeviceAddress;
    private List<Message> mMessages = new LinkedList<>();

    private TextView mMessageBarTextView;
    private TextView mTemperatureTextView;
    private ImageButton mTemperatureButton;
    private Switch mAirPurifierSwitch;
    private RecyclerView.Adapter mMessageRecyclerViewAdapter;

    private BluetoothManager mBluetoothManager;
    private BluetoothAdapter mBluetoothAdapter;
    private BluetoothGatt mBluetoothGatt;
    private BroadcastReceiver mPairRequestReceiver;
    private BroadcastReceiver mBondStateReceiver;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_device_operator);

        mTargetDeviceAddress = getIntent().getStringExtra(getString(R.string.intent_target_address_key));
        mPolicyStorage = new PrivacyPolicyStorageImpl(getContentResolver(), mTargetDeviceAddress);
        PRIVACY_POLICY_ACCEPT_UUID = UUID.fromString(getString(R.string.uuid_accept_privacy_policy));

        initView();
        initBluetooth();
    }

    private void initView() {
        mMessageBarTextView = (TextView) findViewById(R.id.text_operation_message_bar);
        mTemperatureTextView = (TextView) findViewById(R.id.edit_operation_current_temperature);
        mTemperatureButton = (ImageButton) findViewById(R.id.button_operation_send_temperature);
        mTemperatureButton.setOnClickListener(mTemperatureClickedListener);

        mAirPurifierSwitch = (Switch) findViewById(R.id.switch_operation_air_purifier);
        mAirPurifierSwitch.setOnClickListener(mOnAirPurifierSwitchCheckedListener);

        RecyclerView mMessageRecyclerView = (RecyclerView) findViewById(R.id.recycler_view_operation_info);
        mMessageRecyclerViewAdapter = new MessageRecyclerViewAdapter(getResources(), mMessages);
        mMessageRecyclerView.setAdapter(mMessageRecyclerViewAdapter);
    }

    private void initBluetooth() {
        mBluetoothManager = (BluetoothManager) getSystemService(Context.BLUETOOTH_SERVICE);
        if (mBluetoothManager != null) {
            mBluetoothAdapter = mBluetoothManager.getAdapter();
            if (mBluetoothAdapter != null)
                connect();
        }
    }

    private final UUID ROOT_UUID = UUID.fromString("01122334-4556-6778-899a-abbccddeeff0");
    private final UUID TEMPERATURE_UUID = UUID.fromString("01122335-4556-6778-899a-abbccddeeff0");
    private final UUID AIR_UUID = UUID.fromString("01122336-4556-6778-899a-abbccddeeff0");
    private final UUID CCCD_UUID = UUID.fromString("00002902-0000-1000-8000-00805f9b34fb");

    private View.OnClickListener mTemperatureClickedListener = new View.OnClickListener() {
        @Override
        public void onClick(View v) {
            if (mBluetoothGatt != null) {
                BluetoothGattService service = mBluetoothGatt.getService(ROOT_UUID);
                BluetoothGattCharacteristic temperature = service == null ? null : service.getCharacteristic(TEMPERATURE_UUID);
                if (temperature != null) {
                    notifyTemperature();
                }
            }
        }
    };

    private View.OnClickListener mOnAirPurifierSwitchCheckedListener = new View.OnClickListener() {
        @Override
        public void onClick(View v) {
            if (mBluetoothGatt != null) {
                BluetoothGattService service = mBluetoothGatt.getService(ROOT_UUID);
                BluetoothGattCharacteristic airPurifier = service == null ? null : service.getCharacteristic(AIR_UUID);
                if (airPurifier != null) {
                    airPurifier.setValue(new byte[]{(byte) (mAirPurifierSwitch.isChecked() ? 1 : 0)});
                    mBluetoothGatt.writeCharacteristic(airPurifier);
                }
            }
        }
    };

    private boolean isSubscribe = false;

    private void notifyTemperature() {
        if (mBluetoothGatt == null) return;
        synchronized (mBluetoothGatt) {
            BluetoothGattService service = mBluetoothGatt.getService(ROOT_UUID);
            BluetoothGattCharacteristic characteristic = service.getCharacteristic(TEMPERATURE_UUID);
            if (characteristic != null) {
                mBluetoothGatt.setCharacteristicNotification(characteristic, !isSubscribe);
                BluetoothGattDescriptor descriptor = characteristic.getDescriptor(CCCD_UUID);
                descriptor.setValue(!isSubscribe ?
                        BluetoothGattDescriptor.ENABLE_NOTIFICATION_VALUE : BluetoothGattDescriptor.DISABLE_NOTIFICATION_VALUE);
                mBluetoothGatt.writeDescriptor(descriptor);
            }
        }
    }

    private void requestAirPurifier() {
        synchronized (mBluetoothGatt) {
            BluetoothGattService service = mBluetoothGatt.getService(ROOT_UUID);
            BluetoothGattCharacteristic characteristic = service.getCharacteristic(AIR_UUID);
            if (characteristic != null) {
                mBluetoothGatt.readCharacteristic(characteristic);
            }
        }
    }

    private void resolveTemperature(final int value) {
        runOnUiThread(new Runnable() {
            @Override
            public void run() {
                mTemperatureTextView.setText(String.valueOf(value));
            }
        });
    }

    private void resolveAirPurifier(final int value) {
        runOnUiThread(new Runnable() {
            @Override
            public void run() {
                mAirPurifierSwitch.setChecked(value == 1);
            }
        });
    }

    private List<Scheduled> mPermissionSettingQueue = new LinkedList<>();

    private BLECustomScript mBLECustomScript = new BLECustomScript() {

        int expectedSuccessValue = 0;
        int successValue = 0;

        void requestServices() {
            if (successValue == expectedSuccessValue) {
                setMessageBar("Try to request services.");
                addMessage("Request Services", "Try to get temperature & air purifier state.");
                notifyTemperature();
            }
        }

        private boolean settingPermission() {
            if (mPermissionSettingQueue.size() > 0) {
                Log.i(TAG, "Continue to work.");
                mPermissionSettingQueue.remove(0).work();
                return false;
            }
            return true;
        }

        @Override
        public void onDiscoverServiceCompleted(final BluetoothGatt gatt) {
            expectedSuccessValue = 0;
            successValue = 0;

            setMessageBar("Enable privacy policy setting.");
            for (BluetoothGattService service : gatt.getServices()) {
                List<String> layerUUIDs = new LinkedList<>();
                layerUUIDs.add(String.valueOf(service.getUuid()));
                //Write service layer selection
                int selection = mPolicyStorage.get(layerUUIDs);
                if (selection != -1) {
                    final BluetoothGattCharacteristic characteristic = service.getCharacteristic(PRIVACY_POLICY_ACCEPT_UUID);
                    if (characteristic != null) {
                        final int finalSelection = selection;
                        mPermissionSettingQueue.add(new Scheduled() {
                            @Override
                            public void work() {
                                setCharacteristicPermission(mBluetoothGatt, characteristic, finalSelection);
                            }
                        });
                        addMessage("Policy Setting", "Set " + BaseUtils.stringList2ArrayString(layerUUIDs) + " to " + selection + ".");
                    }
                }
                for (BluetoothGattCharacteristic characteristic2 : service.getCharacteristics()) {
                    layerUUIDs.add(String.valueOf(characteristic2.getUuid()));
                    //Write characteristic layer selection
                    selection = mPolicyStorage.get(layerUUIDs);
                    if (selection != -1) {
                        final BluetoothGattDescriptor descriptor = characteristic2.getDescriptor(PRIVACY_POLICY_ACCEPT_UUID);
                        if (descriptor != null) {
                            final int finalSelection1 = selection;
                            mPermissionSettingQueue.add(new Scheduled() {
                                @Override
                                public void work() {
                                    setDescriptorPermission(gatt, descriptor, finalSelection1);
                                }
                            });
                            addMessage("Policy Setting", "Set " + BaseUtils.stringList2ArrayString(layerUUIDs) + " to " + selection + ".");
                        }
                    }
                    layerUUIDs.remove(layerUUIDs.size() - 1);
                }
            }
            settingPermission();
        }

        private void setCharacteristicPermission(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int selection) {
            characteristic.setValue(new byte[]{(byte) selection});
            Log.i(TAG, String.valueOf(selection));
            gatt.writeCharacteristic(characteristic);
            expectedSuccessValue++;
        }

        private void setDescriptorPermission(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int selection) {
            descriptor.setValue(new byte[]{(byte) selection});
            Log.i(TAG, String.valueOf(selection));
            gatt.writeDescriptor(descriptor);
            expectedSuccessValue++;
        }

        @Override
        public void onWriteCharacteristicCompleted(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status, byte[] values) {
            if (PRIVACY_POLICY_ACCEPT_UUID.equals(characteristic.getUuid())) {
                successValue++;
                List<String> layerUUIDs = new LinkedList<>();
                layerUUIDs.add(characteristic.getService().getUuid().toString());
                addMessage("Policy Setting",
                        "Set " + BaseUtils.stringList2ArrayString(layerUUIDs) + (status == GATT_SUCCESS ? " success" : " failure") + " .");
                if (settingPermission())
                    requestServices();
            } else if (AIR_UUID.equals(characteristic.getUuid())) {
                addMessage("Air Purifier Setting", "Try to turn " + (BaseUtils.bytesToInt(values) == 1 ? "on " : "off ") + "the air purifier" + ".");
                requestAirPurifier();
            }
        }

        @Override
        public void onReadCharacteristicCompleted(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status, byte[] values) {
            if (AIR_UUID.equals(characteristic.getUuid())) {
                resolveAirPurifier(BaseUtils.bytesToInt(values));
                addMessage("Air Purifier Setting", "The air purifier is " + (BaseUtils.bytesToInt(values) == 1 ? "on" : "off") + ".");
            }
        }

        @Override
        public void onWriteDescriptorCompleted(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status, byte[] values) {
            if (PRIVACY_POLICY_ACCEPT_UUID.equals(descriptor.getUuid())) {
                successValue++;
                List<String> layerUUIDs = new LinkedList<>();
                layerUUIDs.add(descriptor.getCharacteristic().getService().getUuid().toString());
                layerUUIDs.add(descriptor.getCharacteristic().getUuid().toString());
                addMessage("Policy Setting", "Set " + BaseUtils.stringList2ArrayString(layerUUIDs) +
                        (status == GATT_SUCCESS ? " success" : " failure") + " .");
                if (settingPermission())
                    requestServices();
            } else if (CCCD_UUID.equals(descriptor.getUuid())) {
                addMessage("Temperature Setting", (!isSubscribe ? "Subscribe" : "Unsubscribe")
                        + " temperature data of air conditioner.");
                isSubscribe = !isSubscribe;
                runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        mTemperatureButton.setImageResource(isSubscribe ? R.drawable.ic_timer_off_black_24dp : R.drawable.ic_timer_black_24dp);
                    }
                });
                requestAirPurifier();
                setMessageBar("Device setting finished.");
            }
        }

        @Override
        public void onReadDescriptorCompleted(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status, byte[] values) {

        }

        @Override
        public void onNotificationReceived(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic) {
            int value = characteristic.getIntValue(BluetoothGattCharacteristic.FORMAT_UINT8, 0);
            resolveTemperature(value);
            addMessage("Temperature Setting", "Receive temperature value : [" + value + "].");
        }

        @Override
        public void onPairRequestReceived(BluetoothDevice device, int pin) {

        }

        @Override
        public void onPaired(BluetoothDevice device) {

        }

        @Override
        public void onUnpaired(BluetoothDevice device) {

        }

    };

    private BluetoothGattCallback mBluetoothGattCallback = new BluetoothGattCallback() {

        @Override
        public void onConnectionStateChange(BluetoothGatt gatt, int status, int newState) {
            String mac = gatt.getDevice().getAddress();
            if (status == GATT_SUCCESS) {
                if (newState == BluetoothGatt.STATE_CONNECTED) {
                    gatt.discoverServices();
                    addMessage("Connection State Change ", "Device " + mac + " is connected.");
                } else if (newState == BluetoothGatt.STATE_DISCONNECTED) {
                    mBluetoothGatt.close();
                    mBluetoothGatt = null;
                    addMessage("Connection State Change ", "Device " + mac + " is disconnected.");
                }
            }
            //重連機制，但好像只會一直bind住。
            else if (status == 133 && newState == BluetoothGatt.STATE_DISCONNECTED) {
                mBluetoothGatt.close();
                connect();
            }
        }

        @Override
        public void onServicesDiscovered(BluetoothGatt gatt, int status) {
            if (status == GATT_SUCCESS) {
                mBLECustomScript.onDiscoverServiceCompleted(gatt);
            }
        }

        @Override
        public void onCharacteristicChanged(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic) {
            mBLECustomScript.onNotificationReceived(gatt, characteristic);
        }

        @Override
        public void onCharacteristicRead(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
//            if ((status & (BluetoothGatt.GATT_INSUFFICIENT_AUTHENTICATION | BluetoothGatt.GATT_INSUFFICIENT_ENCRYPTION)) != 0) {
//                boolean isSuccessful = gatt.getDevice().createBond();
//                Log.i(TAG, "Create Bond is " + isSuccessful);
//            }
            mBLECustomScript.onReadCharacteristicCompleted(mBluetoothGatt, characteristic, status, characteristic.getValue());

        }

        @Override
        public void onCharacteristicWrite(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
            mBLECustomScript.onWriteCharacteristicCompleted(mBluetoothGatt, characteristic, status, characteristic.getValue());
        }


        @Override
        public void onDescriptorRead(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
            mBLECustomScript.onReadDescriptorCompleted(mBluetoothGatt, descriptor, status, descriptor.getValue());
        }

        @Override
        public void onDescriptorWrite(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
            mBLECustomScript.onWriteDescriptorCompleted(mBluetoothGatt, descriptor, status, descriptor.getValue());
        }

    };

    private void registerPairRequestReceiver() {
        IntentFilter intentFilter = new IntentFilter(BluetoothDevice.ACTION_PAIRING_REQUEST);
        BroadcastReceiver receiver = new BroadcastReceiver() {
            @Override
            public void onReceive(Context context, Intent intent) {
                if (intent.getAction().equals(BluetoothDevice.ACTION_PAIRING_REQUEST)) {
                    Log.d(TAG, "PIN " + intent.getIntExtra(BluetoothDevice.EXTRA_PAIRING_KEY, 0));
                    BluetoothDevice device = intent.getParcelableExtra(BluetoothDevice.EXTRA_DEVICE);
                    int pin = intent.getIntExtra(BluetoothDevice.EXTRA_PAIRING_KEY, 0);
                    mBLECustomScript.onPairRequestReceived(device, pin);
                }
            }
        };
        registerReceiver(receiver, intentFilter);
        mPairRequestReceiver = receiver;
    }

    private void unregisterPairRequestReceiver() {
        if (mPairRequestReceiver != null) {
            unregisterReceiver(mPairRequestReceiver);
            mPairRequestReceiver = null;
        }
    }

    private void registerBondStateReceiver() {
//        mMessageCallback.displayMessage("Pair", "Ready to pair " + mTargetAddress, true);
        IntentFilter intentFilter = new IntentFilter(BluetoothDevice.ACTION_BOND_STATE_CHANGED);
        BroadcastReceiver receiver = new BroadcastReceiver() {
            @Override
            public void onReceive(Context context, Intent intent) {
                String action = intent.getAction();
                if (BluetoothDevice.ACTION_BOND_STATE_CHANGED.equals(action)) {
                    final int state = intent.getIntExtra(BluetoothDevice.EXTRA_BOND_STATE, BluetoothDevice.ERROR);
                    final int preState = intent.getIntExtra(BluetoothDevice.EXTRA_PREVIOUS_BOND_STATE, BluetoothDevice.ERROR);
                    Log.i(TAG, "Bond [" + preState + " to " + state + "].");
                    BluetoothDevice device = intent.getParcelableExtra(BluetoothDevice.EXTRA_DEVICE);
                    if (state == BluetoothDevice.BOND_BONDED && preState == BluetoothDevice.BOND_BONDING) {
                        mBLECustomScript.onPaired(device);
                    } else if (state == BluetoothDevice.BOND_NONE && preState == BluetoothDevice.BOND_BONDED) {
                        mBLECustomScript.onUnpaired(device);
                    }
                }

            }
        };
        registerReceiver(receiver, intentFilter);
        mBondStateReceiver = receiver;
    }

    private void unregisterBondStateReceiver() {
        if (mBondStateReceiver != null) {
            unregisterReceiver(mBondStateReceiver);
            mBondStateReceiver = null;
        }
    }

    private boolean connect() {
        if (mBluetoothAdapter == null) return false;
        BluetoothDevice device = mBluetoothAdapter.getRemoteDevice(mTargetDeviceAddress);

        int connectionState = mBluetoothManager.getConnectionState(device, BluetoothProfile.GATT);
        Log.i(TAG, "Device connection status is " + connectionState + ".");

        if (device != null && connectionState == BluetoothProfile.STATE_DISCONNECTED) {
            mBluetoothGatt = device.connectGatt(this, false, mBluetoothGattCallback);
            return mBluetoothGatt != null;
        } else {
            return false;
        }
    }


    private void disconnect() {
        if (mBluetoothGatt != null) {
            int connectionState = mBluetoothManager.getConnectionState(mBluetoothGatt.getDevice(), BluetoothProfile.GATT);
            Log.i(TAG, "Device connection status is " + connectionState + ".");
            if (connectionState == BluetoothProfile.STATE_CONNECTED) {
//                boolean isBonded = mBluetoothAdapter.getBondedDevices().contains(mBluetoothGatt.getDevice());
//                Log.i(TAG, "Device is already bonded. :" + isBonded);
//                try {
//                    Method method = mBluetoothGatt.getDevice().getClass().getMethod("removeBond", (Class[]) null);
//                    method.invoke(mBluetoothGatt.getDevice(), (Object[]) null);
//                } catch (NoSuchMethodException | InvocationTargetException | IllegalAccessException e) {
//                    e.printStackTrace();
//                }
                mBluetoothGatt.disconnect();
            }
        }
    }

    private void setMessageBar(final String message) {
        runOnUiThread(new Runnable() {
            @Override
            public void run() {
                mMessageBarTextView.setText(message);
            }
        });
    }

    private void addMessage(String event, String message) {
        addMessage(event, message, false);
    }

    private void addMessage(final String event, final String message, final boolean isMarked) {
        runOnUiThread(new Runnable() {
            @Override
            public void run() {
                mMessages.add(new Message()
                        .setEvent(event)
                        .setContent(message)
                        .setMarked(isMarked));
                mMessageRecyclerViewAdapter.notifyItemInserted(mMessages.size() - 1);
            }
        });
    }

    @Override
    public void onBackPressed() {
        super.onBackPressed();
        if (mBluetoothGatt != null) {
            disconnect();
            mBluetoothGatt.close();
            mBluetoothGatt = null;
        }
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();
        if (mBluetoothGatt != null) {
            disconnect();
            mBluetoothGatt.close();
            mBluetoothGatt = null;
        }
    }
}
